"""
The base module provides classes wrapping shared libraries.
"""
import ctypes
import ctypes.util
import os

# Import types directly, so they don't have to be prefixed with "ctypes.".
from ctypes import c_char_p, c_int, c_uint, c_void_p, c_long, c_int32, c_uint16

import evemu.exception


def raise_error_if(raise_error, result, func, args):
    """
    Raise an ExecutionError for an unexpected result (raise_error == True).

    The exception message includes the API call (name) plus arguments, the
    unexpected result and, if errno is not zero, text describing the
    error number.
    """
    def get_call_str():
        """ Returns a str 'function_name(argument_values...)'. """
        strargs = []
        for (num, arg) in enumerate(func.argtypes):
            # convert args to str for readable output
            if arg == c_char_p:
                strargs.append('"%s"' % args[num].decode("iso8859-1"))
            elif arg == c_void_p:
                strargs.append(hex(int(args[num])))
            else:
                strargs.append(str(args[num]))
        return "%s(%s)" % (func.__name__, ", ".join(strargs))

    def get_retval_str():
        """ Returns a str with the unexpected return value. """
        return ", Unexpected return value: %s" % result

    def get_errno_str():
        """ Returns a str describing the error number or an empty string. """
        errno = ctypes.get_errno()
        if errno != 0:
            return ", errno[%d]: %s" % (errno, os.strerror(errno))
        else:
            return ""

    if raise_error:
        msg = "%s%s%s" % (get_call_str(), get_retval_str(), get_errno_str())
        raise evemu.exception.ExecutionError(msg)
    else:
        # If the errcheck function returns the argument tuple it receives
        # unchanged, ctypes continues the normal processing it does on the
        # output parameters.
        return args


def expect_eq_zero(result, func, args):
    """ Expect 'result' being equal to zero. """
    return raise_error_if(result != 0, result, func, args)


def expect_ge_zero(result, func, args):
    """ Expect 'result' being greater or equal to zero. """
    return raise_error_if(result < 0, result, func, args)


def expect_gt_zero(result, func, args):
    """ Expect 'result' being greater then zero. """
    return raise_error_if(result <= 0, result, func, args)


def expect_not_none(result, func, args):
    """ Expect 'result' being not None. """
    return raise_error_if(result is None, result, func, args)


class LibraryWrapper(object):
    """
    Base class for wrapping a shared library.
    """
    _loaded_lib = None
        # Class variable containing the instance returned by CDLL(), which
        # represents the shared library.
        # Initialized once, shared between all instances of this class.

    def __init__(self):
        super(LibraryWrapper, self).__init__()
        self._load()

    # Prototypes for the API calls to wrap. Needs to be overwritten by sub
    # classes.
    _api_prototypes = {
        #"API_CALL_NAME": {
        #    "argtypes": sequence of ARGUMENT TYPES,
        #    "restype": RETURN TYPE,
        #    "errcheck": callback for return value checking, optional
        #    },
        }

    @classmethod
    def _load(cls):
        """
        Returns an instance of the wrapped shared library.

        If not already initialized: set argument and return types on API
        calls and optionally a callback function for return value checking.
        Add the API call as attribute to the class at the end.
        """
        if cls._loaded_lib is not None:
            # Already initialized, just return it.
            return cls._loaded_lib

        # Get an instance of the wrapped shared library.
        cls._loaded_lib = cls._cdll()

        # Iterate the API call prototypes.
        for (name, attrs) in cls._api_prototypes.items():
            # Get the API call.
            api_call = getattr(cls._loaded_lib, name)
            # Add argument and return types.
            api_call.argtypes = attrs["argtypes"]
            api_call.restype = attrs["restype"]
            # Optionally, add a callback for return value checking.
            if "errcheck" in attrs:
                api_call.errcheck = attrs["errcheck"]
            # Add the API call as attribute to the class.
            setattr(cls, name, api_call)

        return cls._loaded_lib

    @staticmethod
    # @abc.abstractmethod - Would be nice here, but it can't be mixed with
    #                       @staticmethod until Python 3.3.
    def _cdll():
        """ Returns a new instance of the wrapped shared library. """
        raise NotImplementedError


class LibC(LibraryWrapper):
    """
    Wrapper for API calls to the C library.
    """

    @staticmethod
    def _cdll():
        return ctypes.CDLL("libc.so.6", use_errno=True)

    _api_prototypes = {
        "fdopen": {
            "argtypes": (c_int, c_char_p),
            "restype": c_void_p,
            "errcheck": expect_not_none
            },
        "fflush": {
            "argtypes": (c_void_p,),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        "rewind": {
            "argtypes": (c_void_p,),
            "restype": None,
            },
        }

class LibEvdev(LibraryWrapper):
    """
    Wrapper for API calls to libevdev
    """

    @staticmethod
    def _cdll():
        return ctypes.CDLL("libevdev.so", use_errno=True)

    _api_prototypes = {
        #const char *libevdev_event_type_get_name(unsigned int type);
        "libevdev_event_type_get_name": {
            "argtypes": (c_uint,),
            "restype": c_char_p
            },
        #int libevdev_event_type_from_name(const char *name);
        "libevdev_event_type_from_name": {
            "argtypes": (c_char_p,),
            "restype": c_int
            },
        #const char *libevdev_event_code_get_name(unsigned int type, unsigned int code);
        "libevdev_event_code_get_name": {
            "argtypes": (c_uint, c_uint,),
            "restype": c_char_p
            },
        #int libevdev_event_code_from_name(unsigned int type, const char *name);
        "libevdev_event_code_from_name": {
            "argtypes": (c_uint, c_char_p,),
            "restype": c_int
            },
        #const char *libevdev_property_get_name(unsigned int prop);
        "libevdev_property_get_name": {
            "argtypes": (c_uint,),
            "restype": c_char_p
            },
        #int libevdev_property_from_name(const char *name);
        "libevdev_property_from_name": {
            "argtypes": (c_char_p,),
            "restype": c_int
            },
        }

class LibEvemu(LibraryWrapper):
    """
    Wrapper for API calls to the evemu library.
    """

    _LIBNAME = "libevemu.so"

    @staticmethod
    def _cdll():
        return ctypes.CDLL(LibEvemu._LIBNAME, use_errno=True)

    _api_prototypes = {
        #struct evemu_device *evemu_new(const char *name);
        "evemu_new": {
            "argtypes": (c_char_p,),
            "restype": c_void_p,
            "errcheck": expect_not_none
            },
        #void evemu_delete(struct evemu_device *dev);
        "evemu_delete": {
            "argtypes": (c_void_p,),
            "restype": None
            },
        #unsigned int evemu_get_version(const struct evemu_device *dev);
        "evemu_get_version": {
            "argtypes": (c_void_p,),
            "restype": c_uint,
            },
        #const char *evemu_get_name(const struct evemu_device *dev);
        "evemu_get_name": {
            "argtypes": (c_void_p,),
            "restype": c_char_p,
            "errcheck": expect_not_none
            },
        #void evemu_set_name(struct evemu_device *dev, const char *name);
        "evemu_set_name": {
            "argtypes": (c_void_p, c_char_p),
            "restype": None
            },
        #unsigned int evemu_get_id_bustype(const struct evemu_device *dev);
        "evemu_get_id_bustype": {
            "argtypes": (c_void_p,),
            "restype": c_uint
            },
        #void evemu_set_id_bustype(struct evemu_device *dev,
        #                          unsigned int bustype);
        "evemu_set_id_bustype": {
            "argtypes": (c_void_p, c_uint),
            "restype": None
            },
        #unsigned int evemu_get_id_vendor(const struct evemu_device *dev);
        "evemu_get_id_vendor": {
            "argtypes": (c_void_p,),
            "restype": c_uint
            },
        #void evemu_set_id_vendor(struct evemu_device *dev,
        #                         unsigned int vendor);
        "evemu_set_id_vendor": {
            "argtypes": (c_void_p, c_uint),
            "restype": None
            },
        #unsigned int evemu_get_id_product(const struct evemu_device *dev);
        "evemu_get_id_product": {
            "argtypes": (c_void_p,),
            "restype": c_uint
            },
        #void evemu_set_id_product(struct evemu_device *dev,
        #                          unsigned int product);
        "evemu_set_id_product": {
            "argtypes": (c_void_p, c_uint),
            "restype": None
            },
        #unsigned int evemu_get_id_version(const struct evemu_device *dev);
        "evemu_get_id_version": {
            "argtypes": (c_void_p,),
            "restype": c_uint
            },
        #void evemu_set_id_version(struct evemu_device *dev,
        #                          unsigned int version);
        "evemu_set_id_version": {
            "argtypes": (c_void_p, c_uint),
            "restype": None
            },
        #int evemu_get_abs_current_value(const struct evemu_device *dev,
        #                                int code);
        "evemu_get_abs_current_value": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #int evemu_get_abs_minimum(const struct evemu_device *dev, int code);
        "evemu_get_abs_minimum": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #void evemu_set_abs_minimum(struct evemu_device *dev, int code,
        #                           int min);
        "evemu_set_abs_minimum": {
            "argtypes": (c_void_p, c_int, c_int),
            "restype": None
            },
        #int evemu_get_abs_maximum(const struct evemu_device *dev, int code);
        "evemu_get_abs_maximum": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #void evemu_set_abs_maximum(struct evemu_device *dev, int code,
        #                           int max);
        "evemu_set_abs_maximum": {
            "argtypes": (c_void_p, c_int, c_int),
            "restype": None
            },
        #int evemu_get_abs_fuzz(const struct evemu_device *dev, int code);
        "evemu_get_abs_fuzz": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #void evemu_set_abs_fuzz(struct evemu_device *dev, int code, int fuzz);
        "evemu_set_abs_fuzz": {
            "argtypes": (c_void_p, c_int, c_int),
            "restype": None,
            },
        #int evemu_get_abs_flat(const struct evemu_device *dev, int code);
        "evemu_get_abs_flat": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #void evemu_set_abs_flat(struct evemu_device *dev, int code, int flat);
        "evemu_set_abs_flat": {
            "argtypes": (c_void_p, c_int, c_int),
            "restype": None
            },
        #int evemu_get_abs_resolution(const struct evemu_device *dev,
        #                             int code);
        "evemu_get_abs_resolution": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #void evemu_set_abs_resolution(struct evemu_device *dev, int code,
        #                              int res);
        "evemu_set_abs_resolution": {
            "argtypes": (c_void_p, c_int, c_int),
            "restype": None
            },
        #int evemu_has_prop(const struct evemu_device *dev, int code);
        "evemu_has_prop": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #int evemu_has_event(const struct evemu_device *dev, int type,
        #                    int code);
        "evemu_has_event": {
            "argtypes": (c_void_p, c_int, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #int evemu_has_bit(const struct evemu_device *dev, int type);
        "evemu_has_bit": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_ge_zero
            },
        #int evemu_extract(struct evemu_device *dev, int fd);
        "evemu_extract": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #int evemu_write(const struct evemu_device *dev, FILE *fp);
        "evemu_write": {
            "argtypes": (c_void_p, c_void_p),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #int evemu_read(struct evemu_device *dev, FILE *fp);
        "evemu_read": {
            "argtypes": (c_void_p, c_void_p),
            "restype": c_int,
            "errcheck": expect_gt_zero
            },
        #int evemu_write_event(FILE *fp, const struct input_event *ev);
        "evemu_write_event": {
            "argtypes": (c_void_p, c_void_p),
            "restype": c_int,
            "errcheck": expect_gt_zero
            },
        #int evemu_create_event(struct input_event *ev, int type, int code,
        #                       int value);
        "evemu_create_event": {
            "argtypes": (c_void_p, c_int, c_int, c_int),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #int evemu_read_event(FILE *fp, struct input_event *ev);
        "evemu_read_event": {
            "argtypes": (c_void_p, c_void_p),
            "restype": c_int
            },
        #int evemu_read_event_realtime(FILE *fp, struct input_event *ev,
        #			      struct timeval *evtime);
        "evemu_read_event_realtime": {
            "argtypes": (c_void_p, c_void_p, c_void_p),
            "restype": c_int,
            "errcheck": expect_gt_zero
            },
        #int evemu_record(FILE *fp, int fd, int ms);
        "evemu_record": {
            "argtypes": (c_void_p, c_int, c_int),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #int evemu_play_one(int fd, const struct input_event *ev);
        "evemu_play_one": {
            "argtypes": (c_int, c_void_p),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #int evemu_play(FILE *fp, int fd);
        "evemu_play": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #int evemu_create(struct evemu_device *dev, int fd);
        "evemu_create": {
            "argtypes": (c_void_p, c_int),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #int evemu_create_managed(struct evemu_device *dev);
        "evemu_create_managed": {
            "argtypes": (c_void_p,),
            "restype": c_int,
            "errcheck": expect_eq_zero
            },
        #void evemu_destroy(struct evemu_device *dev);
        "evemu_destroy": {
            "argtypes": (c_void_p,),
            "restype": None
            },
        }

class InputEvent(ctypes.Structure):
    _fields_ = [("sec", c_long),
		("usec", c_long),
		("type", c_uint16),
		("code", c_uint16),
		("value", c_int32)]



class EvEmuError(Exception):
    pass


class WrapperError(EvEmuError):
    pass


class ExecutionError(EvEmuError):
    pass


class TestError(EvEmuError):
    pass


class NullFileHandleError(EvEmuError):
    pass


class SkipTest(Exception):
    pass



# UNUSED AND DEPRECATED. DO NOT EDIT
#
# This file won't be updated and is left here merely to avoid breakage
# in old code.
LIB = "libevemu.so"
ENCODING="iso8859-1"
UINPUT_NODE = "/dev/uinput"

event_types = {
    "EV_SYN": 0x00,
    "EV_KEY": 0x01,
    "EV_REL": 0x02,
    "EV_ABS": 0x03,
    "EV_MSC": 0x04,
    "EV_SW": 0x05,
    "EV_LED": 0x11,
    "EV_SND": 0x12,
    "EV_REP": 0x14,
    "EV_FF": 0x15,
    "EV_PWR": 0x16,
    "EV_FF_STATUS": 0x17,
    "EV_MAX": 0x1f,
    }
event_types["EV_CNT"] = event_types["EV_MAX"] + 1,

event_names = {
    "EV_SYN": "Sync",
    "EV_KEY": "Keys or Buttons",
    "EV_REL": "Relative Axes",
    "EV_ABS": "Absolute Axes",
    "EV_MSC": "Miscellaneous",
    "EV_SW": "Switches",
    "EV_LED": "Leds",
    "EV_SND": "Sound",
    "EV_REP": "Repeat",
    "EV_FF": "Force Feedback",
    "EV_PWR": "Power Management",
    "EV_FF_STATUS": "Force Feedback Status",
}

absolute_axes = {
    "ABS_X": 0x00,
    "ABS_Y": 0x01,
    "ABS_Z": 0x02,
    "ABS_RX": 0x03,
    "ABS_RY": 0x04,
    "ABS_RZ": 0x05,
    "ABS_THROTTLE": 0x06,
    "ABS_RUDDER": 0x07,
    "ABS_WHEEL": 0x08,
    "ABS_GAS": 0x09,
    "ABS_BRAKE": 0x0a,
    "ABS_HAT0X": 0x10,
    "ABS_HAT0Y": 0x11,
    "ABS_HAT1X": 0x12,
    "ABS_HAT1Y": 0x13,
    "ABS_HAT2X": 0x14,
    "ABS_HAT2Y": 0x15,
    "ABS_HAT3X": 0x16,
    "ABS_HAT3Y": 0x17,
    "ABS_PRESSURE": 0x18,
    "ABS_DISTANCE": 0x19,
    "ABS_TILT_X": 0x1a,
    "ABS_TILT_Y": 0x1b,
    "ABS_TOOL_WIDTH": 0x1c,
    "ABS_VOLUME": 0x20,
    "ABS_MISC": 0x28,
    "ABS_MT_SLOT": 0x2f,  # MT slot being modified
    "ABS_MT_TOUCH_MAJOR": 0x30,  # Major axis of touching ellipse
    "ABS_MT_TOUCH_MINOR": 0x31,  # Minor axis (omit if circular)
    "ABS_MT_WIDTH_MAJOR": 0x32,  # Major axis of approaching ellipse
    "ABS_MT_WIDTH_MINOR": 0x33,  # Minor axis (omit if circular)
    "ABS_MT_ORIENTATION": 0x34,  # Ellipse orientation
    "ABS_MT_POSITION_X": 0x35,  # Center X ellipse position
    "ABS_MT_POSITION_Y": 0x36,  # Center Y ellipse position
    "ABS_MT_TOOL_TYPE": 0x37,  # Type of touching device
    "ABS_MT_BLOB_ID": 0x38,  # Group a set of packets as a blob
    "ABS_MT_TRACKING_ID": 0x39,  # Unique ID of initiated contact
    "ABS_MT_PRESSURE": 0x3a,  # Pressure on contact area
    "ABS_MT_DISTANCE": 0x3b, # Contact hover distance
    "ABS_MAX": 0x3f,
    }
# XXX ABS_CNT doesn't always give the same value from test data; disabling it
# for now.
#absolute_axes["ABS_CNT"] = absolute_axes["ABS_MAX"] + 1

buttons = {
    "BTN_MISC": 0x100,
    "BTN_0": 0x100,
    "BTN_1": 0x101,
    "BTN_2": 0x102,
    "BTN_3": 0x103,
    "BTN_4": 0x104,
    "BTN_5": 0x105,
    "BTN_6": 0x106,
    "BTN_7": 0x107,
    "BTN_8": 0x108,
    "BTN_9": 0x109,

    "BTN_MOUSE": 0x110,
    "BTN_LEFT": 0x110,
    "BTN_RIGHT": 0x111,
    "BTN_MIDDLE": 0x112,
    "BTN_SIDE": 0x113,
    "BTN_EXTRA": 0x114,
    "BTN_FORWARD": 0x115,
    "BTN_BACK": 0x116,
    "BTN_TASK": 0x117,

    "BTN_JOYSTICK": 0x120,
    "BTN_TRIGGER": 0x120,
    "BTN_THUMB": 0x121,
    "BTN_THUMB2": 0x122,
    "BTN_TOP": 0x123,
    "BTN_TOP2": 0x124,
    "BTN_PINKIE": 0x125,
    "BTN_BASE": 0x126,
    "BTN_BASE2": 0x127,
    "BTN_BASE3": 0x128,
    "BTN_BASE4": 0x129,
    "BTN_BASE5": 0x12a,
    "BTN_BASE6": 0x12b,
    "BTN_DEAD": 0x12f,

    "BTN_GAMEPAD": 0x130,
    "BTN_A": 0x130,
    "BTN_B": 0x131,
    "BTN_C": 0x132,
    "BTN_X": 0x133,
    "BTN_Y": 0x134,
    "BTN_Z": 0x135,
    "BTN_TL": 0x136,
    "BTN_TR": 0x137,
    "BTN_TL2": 0x138,
    "BTN_TR2": 0x139,
    "BTN_SELECT": 0x13a,
    "BTN_START": 0x13b,
    "BTN_MODE": 0x13c,
    "BTN_THUMBL": 0x13d,
    "BTN_THUMBR": 0x13e,

    "BTN_DIGI": 0x140,
    "BTN_TOOL_PEN": 0x140,
    "BTN_TOOL_RUBBER": 0x141,
    "BTN_TOOL_BRUSH": 0x142,
    "BTN_TOOL_PENCIL": 0x143,
    "BTN_TOOL_AIRBRUSH": 0x144,
    "BTN_TOOL_FINGER": 0x145,
    "BTN_TOOL_MOUSE": 0x146,
    "BTN_TOOL_LENS": 0x147,
    "BTN_TOUCH": 0x14a,
    "BTN_STYLUS": 0x14b,
    "BTN_STYLUS2": 0x14c,
    "BTN_TOOL_DOUBLETAP": 0x14d,
    "BTN_TOOL_TRIPLETAP": 0x14e,
    "BTN_TOOL_QUADTAP": 0x14f, # Four fingers on trackpad

    "BTN_WHEEL": 0x150,
    "BTN_GEAR_DOWN": 0x150,
    "BTN_GEAR_UP": 0x151,

    "BTN_TRIGGER_HAPPY": 0x2c0,
    "BTN_TRIGGER_HAPPY1": 0x2c0,
    "BTN_TRIGGER_HAPPY2": 0x2c1,
    "BTN_TRIGGER_HAPPY3": 0x2c2,
    "BTN_TRIGGER_HAPPY4": 0x2c3,
    "BTN_TRIGGER_HAPPY5": 0x2c4,
    "BTN_TRIGGER_HAPPY6": 0x2c5,
    "BTN_TRIGGER_HAPPY7": 0x2c6,
    "BTN_TRIGGER_HAPPY8": 0x2c7,
    "BTN_TRIGGER_HAPPY9": 0x2c8,
    "BTN_TRIGGER_HAPPY10": 0x2c9,
    "BTN_TRIGGER_HAPPY11": 0x2ca,
    "BTN_TRIGGER_HAPPY12": 0x2cb,
    "BTN_TRIGGER_HAPPY13": 0x2cc,
    "BTN_TRIGGER_HAPPY14": 0x2cd,
    "BTN_TRIGGER_HAPPY15": 0x2ce,
    "BTN_TRIGGER_HAPPY16": 0x2cf,
    "BTN_TRIGGER_HAPPY17": 0x2d0,
    "BTN_TRIGGER_HAPPY18": 0x2d1,
    "BTN_TRIGGER_HAPPY19": 0x2d2,
    "BTN_TRIGGER_HAPPY20": 0x2d3,
    "BTN_TRIGGER_HAPPY21": 0x2d4,
    "BTN_TRIGGER_HAPPY22": 0x2d5,
    "BTN_TRIGGER_HAPPY23": 0x2d6,
    "BTN_TRIGGER_HAPPY24": 0x2d7,
    "BTN_TRIGGER_HAPPY25": 0x2d8,
    "BTN_TRIGGER_HAPPY26": 0x2d9,
    "BTN_TRIGGER_HAPPY27": 0x2da,
    "BTN_TRIGGER_HAPPY28": 0x2db,
    "BTN_TRIGGER_HAPPY29": 0x2dc,
    "BTN_TRIGGER_HAPPY30": 0x2dd,
    "BTN_TRIGGER_HAPPY31": 0x2de,
    "BTN_TRIGGER_HAPPY32": 0x2df,
    "BTN_TRIGGER_HAPPY33": 0x2e0,
    "BTN_TRIGGER_HAPPY34": 0x2e1,
    "BTN_TRIGGER_HAPPY35": 0x2e2,
    "BTN_TRIGGER_HAPPY36": 0x2e3,
    "BTN_TRIGGER_HAPPY37": 0x2e4,
    "BTN_TRIGGER_HAPPY38": 0x2e5,
    "BTN_TRIGGER_HAPPY39": 0x2e6,
    "BTN_TRIGGER_HAPPY40": 0x2e7,
    }



"""
The evemu module provides the Python interface to the kernel-level input device
raw events.
"""

# Copyright 2011-2012 Canonical Ltd.
# Copyright 2014 Red Hat, Inc.
#
# This library is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ctypes
import glob
import os
import re
import stat
import tempfile

import evemu.base

__all__ = ["Device",
           "InputEvent",
           "event_get_value",
           "event_get_name",
           "input_prop_get_value",
           "input_prop_get_name"]

_libevdev = evemu.base.LibEvdev()

def event_get_value(event_type, event_code = None):
    """
    Return the integer-value for the given event type and/or code string
    e.g. "EV_ABS" returns 0x03, ("EV_ABS", "ABS_Y") returns 0x01.
    Unknown event types or type/code combinations return None.

    If an event code is passed, the event type may be given as integer or
    string.
    """
    t = -1
    c = -1

    if isinstance(event_type, int):
        event_type = _libevdev.libevdev_event_type_get_name(event_type)
        if event_type is None:
            return None

        event_type = event_type.decode("iso8859-1")

    event_type = str(event_type).encode("iso8859-1")
    t = _libevdev.libevdev_event_type_from_name(event_type)

    if event_code is None:
        return None if t < 0 else t

    if isinstance(event_code, int):
        event_code = _libevdev.libevdev_event_code_get_name(t, event_code)
        if event_code is None:
            return None

        event_code = event_code.decode("iso8859-1")

    event_code = str(event_code).encode("iso8859-1")
    c = _libevdev.libevdev_event_code_from_name(t, event_code)

    return None if c < 0 else c

def event_get_name(event_type, event_code = None):
    """
    Return the string-value for the given event type and/or code value
    e.g. 0x03 returns "EV_ABS", ("EV_ABS", 0x01) returns "ABS_Y"
    Unknown event types or type/code combinations return None.

    If an event code is passed, the event type may be given as integer or
    string.
    """
    if not isinstance(event_type, int):
        event_type = event_get_value(event_type)

    if event_type is None:
        return None

    if event_code is None:
        type_name = _libevdev.libevdev_event_type_get_name(event_type)

        if type_name is None:
            return None

        return type_name.decode("iso8859-1")

    if not isinstance(event_code, int):
        event_code = event_get_value(event_type, event_code)

    if event_code is None:
        return None

    code_name = _libevdev.libevdev_event_code_get_name(event_type, event_code)
    if code_name is None:
        return None

    return code_name.decode("iso8859-1")

def input_prop_get_name(prop):
    """
    Return the name of the input property, or None if undefined.
    """
    if not isinstance(prop, int):
        prop = input_prop_get_value(prop)

    if prop is None:
        return None

    prop = _libevdev.libevdev_property_get_name(prop)
    if prop is None:
        return None

    return prop.decode("iso8859-1")

def input_prop_get_value(prop):
    """
    Return the value of the input property, or None if undefined.
    """
    if isinstance(prop, int):
        prop = input_prop_get_name(prop)

    if prop is None:
        return None

    prop = str(prop).encode("iso8859-1")
    prop = _libevdev.libevdev_property_from_name(prop)
    return None if prop < 0 else prop

class InputEvent(object):
    __slots__ = 'sec', 'usec', 'type', 'code', 'value'

    def __init__(self, sec, usec, type, code, value):
        self.sec = sec
        self.usec = usec
        self.type = type
        self.code = code
        self.value = value

    def matches(self, type, code = None):
        """
        If code is None, return True if the event matches the given event
        type. If code is not None, return True if the event matches the
        given type/code pair.

        type and code may be ints or string-like ("EV_ABS", "ABS_X").
        """
        if event_get_value(type) != self.type:
            return False

        if code != None and event_get_value(self.type, code) != self.code:
            return False

        return True

    def __str__(self):
        f = tempfile.TemporaryFile()
        libc = evemu.base.LibC()
        fp = libc.fdopen(f.fileno(), b"w+")

        event = evemu.base.InputEvent()
        event.sec = self.sec
        event.usec = self.usec
        event.type = self.type
        event.code = self.code
        event.value = self.value

        libevemu = evemu.base.LibEvemu()
        libevemu.evemu_write_event(fp, ctypes.byref(event))
        libc.fflush(fp)
        f.seek(0)
        return f.readline().rstrip()

class Device(object):
    """
    Encapsulates a raw kernel input event device, either an existing one as
    reported by the kernel or a pseudodevice as created through a .prop file.
    """

    def __init__(self, f, create=True):
        """
        Initialize an evemu Device.

        args:
        f -- a file object or filename string for either an existing input
        device node (/dev/input/eventNN) or an evemu prop file that can be used
        to create a pseudo-device node.
        create -- If f points to an evemu prop file, 'create' specifies if a
        uinput device should be created
        """

        if type(f) == str:
            self._file = open(f)
        elif hasattr(f, "read"):
            self._file = f
        else:
            raise TypeError("expected file or file name")

        self._is_propfile = self._check_is_propfile(self._file)
        self._libc = evemu.base.LibC()
        self._libevemu = evemu.base.LibEvemu()

        self._evemu_device = self._libevemu.evemu_new(b"")

        if self._is_propfile:
            fs = self._libc.fdopen(self._file.fileno(), b"r")
            self._libevemu.evemu_read(self._evemu_device, fs)
            if create:
                self._file = self._create_devnode()
        else:
            self._libevemu.evemu_extract(self._evemu_device,
                                         self._file.fileno())

    def __del__(self):
        if hasattr(self, "_is_propfile") and self._is_propfile:
            self._file.close()
            self._libevemu.evemu_destroy(self._evemu_device)

    def _create_devnode(self):
        self._libevemu.evemu_create_managed(self._evemu_device)
        return open(self._find_newest_devnode(self.name), 'r+b', buffering=0)

    def _find_newest_devnode(self, target_name):
        newest_node = (None, float(0))
        for sysname in glob.glob("/sys/class/input/event*/device/name"):
            with open(sysname) as f:
                name = f.read().rstrip()
                if name == target_name:
                    ev = re.search("(event\d+)", sysname)
                    if ev:
                       devname = os.path.join("/dev/input", ev.group(1))
                       ctime = os.stat(devname).st_ctime
                       if ctime > newest_node[1]:
                           newest_node = (devname, ctime)
        return newest_node[0]

    def _check_is_propfile(self, f):
        if stat.S_ISCHR(os.fstat(f.fileno()).st_mode):
            return False

        result = False
        for line in f.readlines():
            if line.startswith("N:"):
                result = True
                break
            elif line.startswith("# EVEMU"):
                result = True
                break
            elif line[0] != "#":
                raise TypeError("file must be a device special or prop file")

        f.seek(0)
        return result

    def describe(self, prop_file):
        """
        Gathers information about the input device and prints it
        to prop_file. This information can be parsed later when constructing
        a Device to create a virtual input device with the same properties.

        You need the required permissions to access the device file to
        succeed (usually root).

        prop_file must be a real file with fileno(), not file-like.
        """
        if not hasattr(prop_file, "fileno"):
            raise TypeError("expected file")

        fs = self._libc.fdopen(prop_file.fileno(), b"w")
        self._libevemu.evemu_write(self._evemu_device, fs)
        self._libc.fflush(fs)

    def events(self, events_file=None):
        """
        Reads the events from the given file and returns them as a list of
        dicts.

        If not None, events_file must be a real file with fileno(), not
        file-like. If None, the file used for creating this device is used.
        """
        if events_file:
            if not hasattr(events_file, "fileno"):
                raise TypeError("expected file")
        else:
            events_file = self._file

        fs = self._libc.fdopen(events_file.fileno(), b"r")
        event = evemu.base.InputEvent()
        while self._libevemu.evemu_read_event(fs, ctypes.byref(event)) > 0:
            yield InputEvent(event.sec, event.usec, event.type, event.code, event.value)

        self._libc.rewind(fs)

    def play(self, events_file):
        """
        Replays an event sequence, as provided by the events_file,
        through the input device. The event sequence must be in
        the form created by the record method.

        You need the required permissions to access the device file to
        succeed (usually root).

        events_file must be a real file with fileno(), not file-like.
        """
        if not hasattr(events_file, "fileno"):
            raise TypeError("expected file")

        fs = self._libc.fdopen(events_file.fileno(), b"r")
        self._libevemu.evemu_play(fs, self._file.fileno())

    def record(self, events_file, timeout=10000):
        """
        Captures events from the input device and prints them to the
        events_file. The events can be parsed by the play method,
        allowing a virtual input device to emit the exact same event
        sequence.

        You need the required permissions to access the device file to
        succeed (usually root).

        events_file must be a real file with fileno(), not file-like.
        """
        if not hasattr(events_file, "fileno"):
            raise TypeError("expected file")

        fs = self._libc.fdopen(events_file.fileno(), b"w")
        self._libevemu.evemu_record(fs, self._file.fileno(), timeout)
        self._libc.fflush(fs)

    @property
    def version(self):
        """
        Gets the version of the evemu library used to create the Device.
        """
        return self._libevemu.evemu_get_version(self._evemu_device)

    @property
    def devnode(self):
        """
        Gets the name of the /dev node of the input device.
        """
        return self._file.name

    @property
    def name(self):
        """
        Gets the name of the input device (as reported by the device).
        """
        result = self._libevemu.evemu_get_name(self._evemu_device)
        return result.decode("iso8859-1")

    @property
    def id_bustype(self):
        """
        Identifies the kernel device bustype.
        """
        return self._libevemu.evemu_get_id_bustype(self._evemu_device)

    @property
    def id_vendor(self):
        """
        Identifies the kernel device vendor.
        """
        return self._libevemu.evemu_get_id_vendor(self._evemu_device)

    @property
    def id_product(self):
        """
        Identifies the kernel device product.
        """
        return self._libevemu.evemu_get_id_product(self._evemu_device)

    @property
    def id_version(self):
        """
        Identifies the kernel device version.
        """
        return self._libevemu.evemu_get_id_version(self._evemu_device)

    def get_abs_current_value(self, event_code):
        """
        Return the current value for the given EV_ABS value.

        event_code may be an int or string-like ("ABS_X").
        """
        if not isinstance(event_code, int):
            event_code = event_get_value("EV_ABS", event_code)
        return self._libevemu.evemu_get_abs_current_value(self._evemu_device,
                                                          event_code)

    def get_abs_minimum(self, event_code):
        """
        Return the axis minimum for the given EV_ABS value.

        event_code may be an int or string-like ("ABS_X").
        """
        if not isinstance(event_code, int):
            event_code = evemu.event_get_value("EV_ABS", event_code)
        return self._libevemu.evemu_get_abs_minimum(self._evemu_device,
                                                    event_code)

    def get_abs_maximum(self, event_code):
        """
        Return the axis maximum for the given EV_ABS value.

        event_code may be an int or string-like ("ABS_X").
        """
        if not isinstance(event_code, int):
            event_code = evemu.event_get_value("EV_ABS", event_code)
        return self._libevemu.evemu_get_abs_maximum(self._evemu_device,
                                                    event_code)

    def get_abs_fuzz(self, event_code):
        """
        Return the abs fuzz for the given EV_ABS value.

        event_code may be an int or string-like ("ABS_X").
        """
        if not isinstance(event_code, int):
            event_code = evemu.event_get_value("EV_ABS", event_code)
        return self._libevemu.evemu_get_abs_fuzz(self._evemu_device,
                                                 event_code)

    def get_abs_flat(self, event_code):
        """
        Return the abs flat for the given EV_ABS value.

        event_code may be an int or string-like ("ABS_X").
        """
        if not isinstance(event_code, int):
            event_code = evemu.event_get_value("EV_ABS", event_code)
        return self._libevemu.evemu_get_abs_flat(self._evemu_device,
                                                 event_code)

    def get_abs_resolution(self, event_code):
        """
        Return the resolution for the given EV_ABS value.

        event_code may be an int or string-like ("ABS_X").
        """
        if not isinstance(event_code, int):
            event_code = evemu.event_get_value("EV_ABS", event_code)
        return self._libevemu.evemu_get_abs_resolution(self._evemu_device,
                                                       event_code)

    # don't change 'event_code' to prop, it breaks API
    def has_prop(self, event_code):
        """
        Return True if the device supports the given input property,
        or False otherwise.

        event_code may be an int or string-like ("INPUT_PROP_DIRECT").
        """
        if not isinstance(event_code, int):
            event_code = evemu.input_prop_get_value(event_code)
        result = self._libevemu.evemu_has_prop(self._evemu_device, event_code)
        return bool(result)

    def has_event(self, event_type, event_code):
        """
        Return True if the device supports the given event type/code
        pair, or False otherwise.

        event_type and event_code may be ints or string-like ("EV_REL",
        "REL_X").
        """
        if not isinstance(event_type, int):
            event_type = evemu.event_get_value(event_type)
        if not isinstance(event_code, int):
            event_code = evemu.event_get_value(event_type, event_code)
        result = self._libevemu.evemu_has_event(self._evemu_device,
                                                event_type,
                                                event_code)
        return bool(result)

